/***************************************************************************
 *   Copyright (C) 2005 by Siegfried Nijssen                               *
 *   snijssen@informatik.uni-freiburg.de                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <cstdlib>
#include <fstream>
#include "ggraph.h"
#include "gdatabase.h"
#include "gdfscode.h"
#include "gextensioncollect.h"
#include "gsubgraphiso.h"
#include "gcanonicaldfscomputer.h"
#include "gtrie.h"
#include "glattice.h"
#include "gresultset.h"
#include <fstream>
#include <map>
#include <cmath>

using namespace std;

struct GSearchQueueData {
  GSearchCode *searchcode;
  vector<GTid> tids;
};

double computeChi2 ( vector<GFrequency> &counters, vector<GFrequency> &totals ) {
  double ret = 0.0;
  double E1, E2;
  double V1, V2;
  GFrequency  allclass = 0, tot = 0;
  for ( int i = 0; i < MAXCLASS; i++ ) {
    allclass += counters[i];
    tot += totals[i]; // perhaps not efficient to redo this every time.
  }
  GFrequency totminall = tot - allclass;
  for ( int i = 0; i < MAXCLASS; i++ ) {
    E1 = ( totals[i] * (double) allclass ) / (double) tot;
    E2 = ( totals[i] * (double) totminall ) / (double) tot;
    V1 = counters[i] - E1;
    V2 = totals[i] - counters[i] - E2;
    if ( E1 != 0.0 )
      ret += V1 * V1 / E1;
    if ( E2 != 0.0 )
      ret += V2 * V2 / E2;
  }
  return ret;
}

double computeIG ( vector<GFrequency> &counters, vector<GFrequency> &totals ) {
  double ret = 0.0;

  double entropy = 0.0;
  GFrequency  allclass = 0, tot = 0;//allcalls is support, tot is data set size
  for ( int i = 0; i < MAXCLASS; i++ ) {
    allclass += counters[i];
    tot += totals[i]; // perhaps not efficient to redo this every time.
  }
  for(int i = 0; i < MAXCLASS; i++){
	if(totals[i] != 0) entropy += -1.0 * ((double)totals[i]/(double)tot) * log2((double)totals[i]/(double)tot);
  }
  GFrequency totminall = tot - allclass;//size of negated pattern
  double covEnt = 0.0;
  double uncovEnt = 0.0;
  for ( int i = 0; i < MAXCLASS; i++ ) {
	double proportion = (double)counters[i]/(double)allclass;
	if(proportion != 0) covEnt += -1.0 * proportion * log2(proportion);
	proportion = (double)(totals[i] - counters[i])/(double)totminall;
	if(proportion != 0) uncovEnt += -1.0 * proportion * log2(proportion);
  }
  ret = entropy - ((double)allclass/(double)tot)*covEnt - ((double)totminall/(double)tot)*uncovEnt;
  return ret;
}

double maxExpChi2 ( vector<GFrequency> &counters, vector<GFrequency> &totals , bool infoGain) {//Dirty because I overload ExpChi2 - should change name
  GFrequency temp = counters[0];
  double V1, V2;
  if(!infoGain){
	  counters[0] = 0;
	  V1 = computeChi2 ( counters, totals );
	  counters[0] = temp;
	  temp = counters[1];
	  counters[1] = 0;
	  V2 = computeChi2 ( counters, totals );
	  counters[1] = temp;
  }else{
	  counters[0] = 0;
	  V1 = computeIG ( counters, totals );
	  counters[0] = temp;
	  temp = counters[1];
	  counters[1] = 0;
	  V2 = computeIG ( counters, totals );
	  counters[1] = temp;
  }
  return max ( V1, V2 );
}

// The algorithm works with two "sets" (or queues, or maps) of patterns.
// The ResultSet is the set of patterns that will eventually be returned to
// the user. The pattern in this set with the lowest chi2 value determines
// the threshold on the search.
// The SearchQueue contains patterns of which we still have to consider 
// refinement. In principle, the best pattern is removed from this queue,
// and its refinements are added back to the queue. By choosing the "best"
// first, we hope that the ResultSet gets good values quickly.
// Once the threshold rises, it's possible that it is no longer necessary
// to consider some patterns that were put into the queue previously.
// clean () removes these patterns.
// In practice, the user can specifify a maximum size on the queue.
// This means that if the queue is full, the algorithm falls back
// to depth-first search, evaluating refinements immediately.
// Please note that during this search, the threshold can rise, and
// the queue can become emptier. As soon as this happens, patterns are 
// inserted in the queue again. In the current implementation, this
// interaction between DFS search and search through the queue may however be 
// suboptimal. The main reason for using a restricted queue size, is to give the user
// a possibility to restrict memory usage, as for every pattern in the
// SearchQueue also vectors of tid's are stored, which can consume lots of
// memory for large databases.
class GSearchQueue:public multimap<double,GSearchQueueData*> {
public: 
    int _size;
    GSearchQueue ():_size ( 0 ) { }
    // reimplementations of insert and erase to make sure that size ()
    // is O(1) and not O(n)
    void insert ( double chi2, GSearchQueueData *data ) {
      multimap<double,GSearchQueueData*>::insert ( make_pair ( chi2, data ) );
      ++_size;
    }
    void erase ( GSearchQueue::iterator it ) {
      multimap<double,GSearchQueueData*>::erase ( it );
      --_size;
    }
    int size () const { return _size; }
    
    void clean ( double minchi2 ) {
      iterator it = begin (), it2;
      while ( it != end () && it->first < minchi2 ) {
        it2 = it;
        ++it;
        erase (it2 );
      }
    }
};

void search ( GResultSet &resultset, GSearchQueue &searchqueue, int maxqueuelength, GDatabase &database, GSearchCode *searchcode, vector<GTid> &tids , bool infoGain) {
  GExtensionCollect collect;
  GSubGraphIso subgraphiso ( &collect, &searchcode->graph );
  vector<GTid> tids2;
  
  for ( int i = 0, size = tids.size (); i < size; ++i )
    if ( subgraphiso.run ( tids[i] ) )
      tids2.push_back ( tids[i] );
          
  for ( map<GToken,GExtensionData,GTokenLt>::iterator i = collect.extensions.begin ();
        i != collect.extensions.end ();
        ++i ) {
    searchcode->push_back ( i->first );
    
    if ( searchcode->isCanonical () ) {
     
      if(!infoGain){ 
	      resultset.insert ( computeChi2 ( i->second.frequency, database.classsizes ), 
				 i->second.frequency,
				 &searchcode->graph );
      }else{
	      resultset.insert ( computeIG ( i->second.frequency, database.classsizes ), 
				 i->second.frequency,
				 &searchcode->graph );
      }
      searchqueue.clean ( resultset.minChi2 () );
        
      double expchi2 = maxExpChi2 ( i->second.frequency, database.classsizes , infoGain);
      
      static int busycnt = 0, cnt = 0, cnt2 = 0;
      if ( busycnt % 100 == 0 )
        cout << busycnt << " considered, " << cnt << " candidates, " << cnt2 << " queued " << resultset.minChi2 () << " minchi2 \r" << flush;
      ++busycnt;
      
      if ( expchi2 >= resultset.minChi2 () ) {
        if ( searchqueue.size () < maxqueuelength ) {
          GSearchQueueData *sqd = new GSearchQueueData;
          sqd->searchcode = searchcode->copy ();
          sqd->tids = tids2; // copy, not swap as possibly multiple copies are required.
          searchqueue.insert ( expchi2, sqd );
          ++cnt2;
        }
        else 
          search ( resultset, searchqueue, maxqueuelength, database, searchcode, tids2, infoGain );
        ++cnt;
      }
          
    }
    
    searchcode->pop_back ();
  }
}

void init ( GResultSet &resultset, GSearchQueue &searchqueue, GDatabase &database , bool infoGain) {
  map<GToken,GTokenDatabaseData,GTokenLt>::iterator i = database.tokendata.begin ();
  
  while ( i != database.tokendata.end () ) {
    GSearchQueueData *gsqd = new GSearchQueueData;
    swap ( gsqd->tids, i->second.tids ); // avoid using too much memory
    gsqd->searchcode = new GDFSCode;
    gsqd->searchcode->setup ( i->first );    
    gsqd->searchcode->push_back ( i->first );    
    if(!infoGain){
	    resultset.insert ( computeChi2 ( i->second.frequency, database.classsizes ), i->second.frequency, &gsqd->searchcode->graph );
    }else{
	    resultset.insert ( computeIG ( i->second.frequency, database.classsizes ), i->second.frequency, &gsqd->searchcode->graph );
    }
    searchqueue.clean ( resultset.minChi2 () );
    searchqueue.insert ( maxExpChi2 ( i->second.frequency, database.classsizes, infoGain ), gsqd ); 
    ++i;
  }  
}

void search ( GResultSet &resultset, GDatabase &database, int maxqueuelength, bool infoGain ) {
  GSearchQueue searchqueue;
  GSearchQueue::iterator it;
  
  init ( resultset, searchqueue, database, infoGain );
  
  while ( !searchqueue.empty () ) {
    it = searchqueue.end ();
    --it;
    GSearchQueueData *gsqd = it->second;
    searchqueue.erase ( it );
    
    search ( resultset, searchqueue, maxqueuelength, database, gsqd->searchcode, gsqd->tids, infoGain );
    
    delete gsqd->searchcode;
    delete gsqd;    
  }
}

int main(int argc, char *argv[])
{
  GDatabase database;
  GResultSet resultset;
  cout << "YAI-Span: Yet Another Implementation of gSpan" << endl
       << "Siegfried Nijssen (2005-2006)" << endl
       << "=============================================" << endl
       << "Opening " << argv[2] << endl;
  database.read ( argv[2], 0 );
  cout << "Opening " << argv[3] << endl;
  database.read ( argv[3], 1 );
  resultset.setSize ( atoi ( argv[1] ) );
  int maxqueuelength = 1000;
  bool infoGain = false;
  if ( argc >= 5 ){
    infoGain = (strcmp( argv[4], "ig") == 0);
    cout << "Using Information Gain for mining: " << infoGain << endl;
  }
  if ( argc >= 6 )
    maxqueuelength = atoi ( argv[5] );
    
  if ( argc == 7 ) 
    resultset.setMinChi2 ( atof ( argv[6] ) );
    
  if ( maxqueuelength < database.tokendata.size () ) {
    cout << "Please set a larger maxqueuelength: all initial tokens must fit within the queue." << endl;
    return EXIT_FAILURE;
  }

  cout << "Searching " << endl;
  search ( resultset, database, maxqueuelength, infoGain );
  cout << endl << "Finished with minchi2 " << resultset.minChi2 () << endl;
  if ( resultset.shrinkcount > 0 ) {
    cout << "The set has been shrinked " << resultset.shrinkcount << " times." << endl;
  }
  
  cout << resultset;
    
  return EXIT_SUCCESS;
}
