#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import tkinter
import tkinter.filedialog
import tkinter.ttk

import tempfile
import networkx as nx
import matplotlib.pyplot as plt
import time
import numpy as np

from datetime import datetime
from threading import Thread
from PIL import Image, ImageTk

import prepep_ml
import prepep_multi_processing
import logging


POPUPS_BY_MOLECULE = {}


def LoadFile(ev):
    global f_name
    f_name = tkinter.filedialog.Open(master, filetypes=[('*.sdf files', '.sdf')]).show()
    if f_name == '':
        return


class EncodingThread(Thread):
    """
    A thread for running the encoding task asynchroniously (ie. without
    locking the gui).
    """
    
    def __init__(self, f_name, clf_list, num_substructures, num_samplings):
        """
        Constructor : initialize the thread with the needed arguments.

        Args:
            fname : the name of the sdf file
            clf_list : the classifiers list (200)
            num_substructures : 100
            num_samplings : 200

        Returns:
            the thread object, ready to be run.
        """
        Thread.__init__(self)

        self._f_name = f_name
        self._clf_list = clf_list

        self._num_substructures = num_substructures
        self._num_samplings = num_samplings

        self._is_processed = False

        self._M_dataset = None
        self._M_compounds = None
        self._dataset_X = None
        self._pred_res = None

    def run(self):  
        """
        Running the encode task, asynchroniously.
        """
        self._M_dataset, self._M_compounds = prepep_ml.open_dataset(self._f_name)

        self._dataset_X = prepep_multi_processing.encode_dataset_mt(self._M_dataset, self._num_substructures, self._num_samplings)

        self._pred_res = prepep_ml.majority_vote_prediction(self._clf_list, self._dataset_X)

        self._is_processed = True


def EncodeFile(ev):

    global M_dataset
    global M_compounds
    global dataset_X
    global pred_res

    listboxMolecules.delete(0, tkinter.END)
    listboxGraphs.delete(0, tkinter.END)

    thread = EncodingThread(f_name, clf_list, num_substructures, num_samplings)
    thread.start()

    encodingBar.start()

    index = 0

    while not thread._is_processed:
        encodingBar["value"] = index
        encodingBar.update()
        index = index + 2
        time.sleep(0.1)

    encodingBar.stop()

    M_dataset = thread._M_dataset
    M_compounds = thread._M_compounds
    dataset_X = thread._dataset_X
    pred_res = thread._pred_res

    for item in range(len(M_dataset)):

        listboxMolecules.insert(tkinter.END, "[" + str(item + 1) + "] : " + str(M_compounds[item].formula))

        is_frequent = pred_res[item]

        if is_frequent:
            
            listboxMolecules.itemconfig(tkinter.END, fg='red')

    bottomLeftLabel['text'] = ''


def Quit(ev):
    global master
    plt.close("all")
    master.destroy()


class CloseTopLevelMoleculeCallable:

    def __init__(self, top_level, molecule_index):

        self._top_level = top_level
        self._molecule_index = molecule_index
    
    def __call__(self):

        popups = POPUPS_BY_MOLECULE[self._molecule_index]

        for popup in popups:
            popup.destroy()

        self._top_level.destroy()

        POPUPS_BY_MOLECULE[self._molecule_index] = []

class CloseTopLevelCallable:
    """
    Une classe "appelable", c'est à dire qui possède une méthode __call__().
    """

    def __init__(self, top_level):
        """
        Crée un nouvel objet "appelable".

        Arguments:
            topLevel {Toplevel} -- la fenêtre qu'il convient de fermer.

        Returns:
            L'objet "appelable", prêt à être utilisé.
        """
        self._top_level = top_level

    def __call__(self):
        """
        Méthode "callback" appelée par le bouton lors du click.

        Ferme la fenêtre.
        """
        self._top_level.destroy()


def MoleculeSelect(evt):

    global graphs_list
    global graphs_freq
    global molecule_index
    global master

    logging.info("processing MoleculeSelect(evt)")

    molecule_index = listboxMolecules.curselection()[0]

    if not molecule_index in POPUPS_BY_MOLECULE:

        POPUPS_BY_MOLECULE[molecule_index] = []

    is_frequent = pred_res[molecule_index]

    if is_frequent:
        bottomLeftLabel['text'] = "Result: frequent hitter"
    else:
        bottomLeftLabel['text'] = "Result: non-frequent hitter"

    listboxGraphs.delete(0, tkinter.END)

    if is_frequent:

        graphs_list, graphs_freq = prepep_ml.molecule_stat(clf_list, molecule_index, dataset_X, num_samplings, num_substructures)

        for item in range(len(graphs_list)):

            listboxGraphs.insert(tkinter.END, '[' + str(item + 1) + '] : ' + str(graphs_list[item].formula))

    bottomRightLabel['text'] = ''

    file = tempfile.NamedTemporaryFile(suffix=".png", delete=True)

    M_compounds[molecule_index].draw(show=False, filename=file.name)

    image = Image.open(file.name)

    size = image.size

    area = image.crop((0, 0, size[0], size[1] - 20))

    fileCropped = tempfile.NamedTemporaryFile(suffix=".png", delete=True)

    area.save(fileCropped)

    topLevel = tkinter.Toplevel(master)
    topLevel.title("molecule [" + str(molecule_index + 1) + "]")

    photo = ImageTk.PhotoImage(file=fileCropped.name, master=master)

    label = tkinter.Label(topLevel, image=photo)
    label.pack()

    text = tkinter.Label(topLevel, text=str(M_compounds[molecule_index].formula))
    text.pack()

    button = tkinter.Button(topLevel, text="Close", command=CloseTopLevelMoleculeCallable(topLevel, molecule_index))
    button.pack()

    master.mainloop()


def GraphSelect(evt):

    global master
    global molecule_index

    logging.info("processing GraphSelect(evt)")
    logging.info("molecule_index is [%s]" % molecule_index)

    graph_index = listboxGraphs.curselection()[0]

    bottomRightLabel['text'] = "Frequency: " + str(graphs_freq[graph_index])

    logging.info("Frequency: [%s]" % (str(graphs_freq[graph_index])))

    file = tempfile.NamedTemporaryFile(suffix=".png", delete=True)

    logging.info("temporary file name is [%s]" % (str(file.name)))

    graphs_list[graph_index].draw(show=False, filename=file.name)

    topLevel = tkinter.Toplevel(master)
    topLevel.title("Subgraph " + str(graph_index + 1) + " [" + str(graphs_list[graph_index].formula) + "]")

    POPUPS_BY_MOLECULE[molecule_index].append(topLevel)

    photo = ImageTk.PhotoImage(file=file.name, master=master)

    label = tkinter.Label(topLevel, image=photo)
    label.pack()

    button = tkinter.Button(topLevel, text="Close", command=CloseTopLevelCallable(topLevel))
    button.pack()

    master.mainloop()


# def UpdateProgressBar(progress):
#    encodingBar["value"] = progress
#    encodingBar.update()
#    bottomLeftLabel['text'] = "Encoding: "+str(progress)+"%"
# print 'encoding',progress,'%'


def load_classifier_folder(ev):

    global clf_list
    global num_samplings
    global num_substructures
    
    directory = tkinter.filedialog.askdirectory()

    if not directory:
        return

    logging.info("directory is [%s]" % (str(directory)))

    prepep_ml.CLASSIFIER_DIRECTORY_PATH = directory

    encoded_training_data = np.load(directory + '/training_data_X.npy')

    num_samplings = encoded_training_data.shape[0]
    num_substructures = encoded_training_data.shape[2]

    clf_list = prepep_ml.learn_classifier(num_samplings)


if __name__ == '__main__':

    now = datetime.now()

    logFile = "prepep_gui-" + now.strftime("%Y%m%d-%H%M%S.%f") + ".log"
    logging.basicConfig(filename=logFile, level=logging.INFO)

    # model parameters:
    num_substructures = 100
    num_samplings = 200

    # learn classifier:
    clf_list = prepep_ml.learn_classifier(num_samplings)

    G = nx.Graph()

    fig = plt.figure(figsize=(1, 1))
    fig.canvas.set_window_title('Visualization')

    master = tkinter.Tk()
    master.title("PrePeP")

    frameTopLeft = tkinter.Frame(master)
    frameTopLeft.grid(row=1, column=1)

    frameTopRight = tkinter.Frame(master)
    frameTopRight.grid(row=1, column=2)

    select_classifiers_folder_btn = tkinter.Button(frameTopLeft, text="Select classifiers folder")
    select_classifiers_folder_btn.bind("<ButtonRelease-1>", load_classifier_folder)
    select_classifiers_folder_btn.pack(side='left')

    topLabelM = tkinter.Label(master, text='Choose molecule:')
    topEntryM = tkinter.Entry(master, width=10)
    topLabelM.grid(row=2, column=1)

    topLabelG = tkinter.Label(master, text='Choose subgraph:')
    topEntryG = tkinter.Entry(master, width=10)
    topLabelG.grid(row=2, column=2)

    frameLeft = tkinter.Frame(master)
    frameLeft.grid(row=3, column=1)

    frameRight = tkinter.Frame(master)
    frameRight.grid(row=3, column=2)

    bottomLeftLabel = tkinter.Label(master, text='')
    bottomLeftEntry = tkinter.Entry(master, width=10)
    bottomLeftLabel.grid(row=4, column=1)

    bottomRightLabel = tkinter.Label(master, text='')
    bottomRightEntry = tkinter.Entry(master, width=10)
    bottomRightLabel.grid(row=4, column=2)

    loadBtn = tkinter.Button(frameTopLeft, text='Open dataset')
    loadBtn.bind("<ButtonRelease-1>", LoadFile)
    loadBtn.pack(side='left')

    encodeBtn = tkinter.Button(frameTopLeft, text='Encoding')
    encodeBtn.bind("<ButtonRelease-1>", EncodeFile)
    encodeBtn.pack(side='right')

    encodingBar = tkinter.ttk.Progressbar(frameTopRight, orient="horizontal", length=150, mode="indeterminate", maximum=100)
    encodingBar.pack(side='left')
    encodingBar.stop()

    quitBtn = tkinter.Button(frameTopRight, text='Quit')
    quitBtn.bind("<Button-1>", Quit)
    quitBtn.pack(side='right')

    listboxMolecules = tkinter.Listbox(frameLeft, selectmode=tkinter.SINGLE, height=20, exportselection=0)
    listboxMolecules.bind('<<ListboxSelect>>', MoleculeSelect)
    listboxMolecules.pack(side="left")

    scrollbarLeft = tkinter.Scrollbar(frameLeft, orient="vertical")
    scrollbarLeft.config(command=listboxMolecules.yview)
    scrollbarLeft.pack(side="right", fill="y")
    listboxMolecules.config(yscrollcommand=scrollbarLeft.set)

    listboxGraphs = tkinter.Listbox(frameRight, selectmode=tkinter.SINGLE, height=20, exportselection=0)
    listboxGraphs.bind('<<ListboxSelect>>', GraphSelect)
    listboxGraphs.pack(side="left")

    scrollbarRight = tkinter.Scrollbar(frameRight, orient="vertical")
    scrollbarRight.config(command=listboxGraphs.yview)
    scrollbarRight.pack(side="right", fill="y")
    listboxGraphs.config(yscrollcommand=scrollbarRight.set)

    master.mainloop()
