#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import math
import time

import prepep_ml
import multiprocessing

import networkx as nx
import numpy as np

class EncodingResult():
    """
       Une classe qui encapsule les resultats du test d'iso-morphisme.
    """

    def __init__(self, classifier_index, molecule_index, substructure_index, isomorph_test_result):
        """
           Constructeur. 

           Crée un objet de la classe EncodingResult
        """
        self.classifier_index = classifier_index
        self.molecule_index = molecule_index
        self.substructure_index = substructure_index
        self.isomorph_test_result = isomorph_test_result


class EncodingDatasetProcess(multiprocessing.Process):
    """
       Une classe de processus d'encodage des datasets.
    """

    def __init__(self, initial_index, target_index, queue, M_list, num_samplings, num_substructures):
        """
           Constructeur.
           
           Crée un nouveau processus.
        """

        multiprocessing.Process.__init__(self)

        self._initial_index = initial_index
        self._target_index = target_index
        self._queue = queue
        self._M_list = M_list
        self._num_samplings = num_samplings
        self._num_substructures = num_substructures
        self._num_molecules_classify = len(M_list)

    def run(self):
        """
           Code executé par le processus.
        """

        logging.info("running thread [%s][%s]" % (self._initial_index, self._target_index))

        for i in range(self._initial_index, self._target_index + 1):

            logging.info("processing classifier [%s]" % (i))

            f = open('classifier/top_sg_'+str(i)+'.gsp', 'r')
            SG_list = []

            for line in f:
                try:
                    if '#' in line:
                        G2 = nx.Graph()
                        # labels2 = {}
                        # node_colors2 = []
                        line = next(f)
                        while 'v' in line:
                            data = line.split(' ')
                            node = np.int(data[1])
                            element = np.int(data[2])
                            # labels2[node] = CHEM_EL[element]['name']
                            # node_colors2.append(CHEM_EL[element]['color'])
                            # G2.add_node(node,label=labels2[node])
                            G2.add_node(node, label=prepep_ml.CHEM_EL[element]['name'])
                            line = next(f)
                        while 'e' in line:
                            data = line.split(' ')
                            vertice1 = np.int(data[1])
                            vertice2 = np.int(data[2])
                            G2.add_edge(vertice1, vertice2, weight=np.int(data[3]))
                            line = next(f)
                        SG_list.append(G2)
                except ValueError:
                    print("Invalid input:", line)

            f.close()

            # Encode the input data:
            for i_molecule in range(self._num_molecules_classify):

                for i_substructure in range(len(SG_list)):

                    isomorph_test_result = prepep_ml.isomorph_test(self._M_list[i_molecule], SG_list[i_substructure])

                    result = EncodingResult(i-1, i_molecule, i_substructure, isomorph_test_result)

                    self._queue.put(result)

        logging.info("end of run() method")

        self._queue.put("EOP") # EOP : END OF PROCESS


def initialize_processes_pool(queue, M_list, thread_count, num_samplings, num_substructures):
    """
        Une fonction qui initialise le pool de processus.
    """

    pool = []

    step_size = int(math.ceil(float(num_samplings) / thread_count))

    logging.info("step_size is [%s]" % (step_size))

    for i in range(1, thread_count + 2):
        
        logging.debug("i is [%s]" % (i))

        initial_index = step_size * (i - 1) + 1

        if initial_index <= num_samplings:

            target_index = initial_index + step_size - 1

            if (target_index > num_samplings):
                target_index = num_samplings

            logging.debug ("initial_index is [%s]" % (initial_index))
            logging.debug ("target_index is [%s]" % (target_index))

            pool.append(EncodingDatasetProcess(initial_index, target_index, queue, M_list, num_samplings, num_substructures))

    return pool


def start_processes_pool(processes_pool):
    """
        Démarre le pool de processus.
    """
    for process in processes_pool:
    
        process.start()


def encode_dataset_mt(M_list, num_substructures, num_samplings):
    """
        Une fonction d'encodage du dataset multi-processus.
    """

    thread_count = multiprocessing.cpu_count()

    logging.info("thread_count is [%s]" % (thread_count))

    begin = time.time()

    #
    # On a besoin d'une queue pour dialoguer avec les processus.
    #
    queue = multiprocessing.Queue()

    processes_pool = initialize_processes_pool(queue, M_list, thread_count, num_samplings, num_substructures)
    
    start_processes_pool(processes_pool)

    eop_count = 0

    pred_X = np.zeros((num_samplings, len(M_list), num_substructures), bool)

    while eop_count < thread_count:

        result = queue.get()

        if (isinstance(result, str) and result == "EOP"):
            logging.info("END OF PROCESS CODE RECEIVED")
            eop_count += 1

        else:
            classifier_index = result.classifier_index
            molecule_index = result.molecule_index
            substructure_index = result.substructure_index
            isomorph_test_result = result.isomorph_test_result

            pred_X[classifier_index][molecule_index][substructure_index] = isomorph_test_result

    for process in processes_pool:
        process.join()

    logging.info("all processes processed !")

    end = time.time()

    logging.info("time is [%s]" % (int(math.ceil(end - begin))))

    return pred_X

if __name__ == "__main__":
    """
        Le main, point d'entrée du programme.
    """

    logging.basicConfig(level=logging.INFO)

    num_samplings = 200
    num_substructures = 100

    clf_list = prepep_ml.learn_classifier(num_samplings)

    M_list, M_list_pybel = prepep_ml.open_dataset("demo2.sdf")
    # M_list, M_list_pybel = prepep_ml.open_dataset("encoding_test.sdf")

    logging.debug("M_list length is [%s]" % (len(M_list)))
    logging.debug("M_list_pybel length is [%s]" % (len(M_list_pybel)))

    dataset_X = encode_dataset_mt(M_list, num_substructures, num_samplings)

    logging.debug("dataset_X.ndim is [%s]" % (dataset_X.ndim))
    logging.debug("dataset_X.shape is [%s]" % (str(dataset_X.shape)))

    #
    # output for demo2.sdf
    #
    # dataset_X.ndim is [ 3 ]
    # dataset_X.shape is [ (200, 4, 100) ]
    #

    pred_res = prepep_ml.majority_vote_prediction(clf_list, dataset_X)

    logging.debug("pred_res.ndim is [%s]" % (pred_res.ndim))
    logging.debug("pred_res is [%s]" % (pred_res))

    #
    # output for demo2.sdf
    #
    # pred_res.ndim is [ 1 ]
    # pred_res is [ [False False  True  True] ]
    #

    for molecule_id in range(len(pred_res)):

        logging.info("processing molecule [%s]" % (molecule_id))

        formula = M_list_pybel[molecule_id].formula

        is_frequent = pred_res[molecule_id]

        if is_frequent:
            logging.info("[%s] is frequent" % (formula))
        else:
            logging.info("[%s] is not frequent" % (formula))

        if is_frequent:

            Mol_list, subgraphs_frequencies = prepep_ml.molecule_stat(clf_list, molecule_id, dataset_X, num_samplings, num_substructures)

            # logging.debug("Mol_list is [" + str(Mol_list) + "]")
            logging.info("Mol_list length is [%s]" % (len(Mol_list)))
            logging.info("subgraphs_frequencies is [%s]" % (subgraphs_frequencies))

            for Mol in Mol_list:
                logging.info(str(Mol.formula))

