#!/usr/bin/env python3

import numpy as np
import sys
import pybel
import networkx as nx
import openbabel
import collections

from sklearn import tree

DEBUG = True

CHEM_EL = {
    1:{'name':'H','color':'yellow'},
    2:{'name':'He','color':'grey'},
    3:{'name':'Li','color':'grey'},
    4:{'name':'Be','color':'grey'},       
    5:{'name':'B','color':'grey'},
    6:{'name':'C','color':'orange'},
    7:{'name':'N','color':'red'},
    8:{'name':'O','color':'green'},
    9:{'name':'F','color':'blue'},
    10:{'name':'Ne','color':'grey'},
    11:{'name':'Na','color':'grey'},
    12:{'name':'Mg','color':'grey'},
    13:{'name':'Al','color':'grey'},
    14:{'name':'Si','color':'grey'},
    15:{'name':'P','color':'grey'},
    16:{'name':'S','color':'violet'},
    17:{'name':'Cl','color':'brown'},
    18:{'name':'Ar','color':'grey'},
    19:{'name':'K','color':'grey'},
    20:{'name':'Ca','color':'grey'},
    21:{'name':'Sc','color':'grey'},
    22:{'name':'Ti','color':'grey'},
    23:{'name':'V','color':'grey'},
    24:{'name':'Cr','color':'grey'},
    25:{'name':'Mn','color':'grey'},
    26:{'name':'Fe','color':'grey'},
    27:{'name':'Co','color':'grey'},
    28:{'name':'Ni','color':'grey'},
    29:{'name':'Cu','color':'grey'},
    30:{'name':'Zn','color':'grey'},
    31:{'name':'Ga','color':'grey'},
    32:{'name':'Ge','color':'grey'},
    33:{'name':'As','color':'grey'},
    34:{'name':'Se','color':'grey'},
    35:{'name':'Br','color':'grey'},
    36:{'name':'Kr','color':'grey'},
    37:{'name':'Rb','color':'grey'},
    38:{'name':'Sr','color':'grey'},
    39:{'name':'Y','color':'grey'},
    40:{'name':'Zr','color':'grey'},
    41:{'name':'Nb','color':'grey'},
    42:{'name':'Mo','color':'grey'},
    43:{'name':'Tc','color':'grey'},
    44:{'name':'Ru','color':'grey'},
    45:{'name':'Rh','color':'grey'},
    46:{'name':'Pd','color':'grey'},
    47:{'name':'Ag','color':'grey'},
    48:{'name':'Cd','color':'grey'},
    49:{'name':'In','color':'grey'},
    50:{'name':'Sn','color':'grey'},
    51:{'name':'Sb','color':'grey'},
    52:{'name':'Te','color':'grey'},
    53:{'name':'I','color':'grey'},
    54:{'name':'Xe','color':'grey'},
    55:{'name':'Cs','color':'grey'},
    56:{'name':'Ba','color':'grey'},
    57:{'name':'Hf','color':'grey'},
    58:{'name':'Ta','color':'grey'},
    59:{'name':'W','color':'grey'},
    60:{'name':'Re','color':'grey'},
    61:{'name':'Os','color':'grey'},
    62:{'name':'Ir','color':'grey'},
    63:{'name':'Pt','color':'grey'},
    64:{'name':'Au','color':'grey'},
    65:{'name':'Hg','color':'grey'},
    66:{'name':'Tl','color':'grey'},
    67:{'name':'Pb','color':'grey'},
    68:{'name':'Bi','color':'grey'},
    69:{'name':'Po','color':'grey'},
    70:{'name':'At','color':'grey'},
    71:{'name':'Rn','color':'grey'},
    72:{'name':'Fr','color':'grey'},
    73:{'name':'Ra','color':'grey'},
    74:{'name':'Pt','color':'grey'},
    75:{'name':'Ac','color':'grey'},
    76:{'name':'La','color':'grey'},
    77:{'name':'U','color':'grey'},
    78:{'name':'Sm','color':'grey'},
    79:{'name':'Ce','color':'grey'},
    80:{'name':'Nd','color':'grey'},
    81:{'name':'Eu','color':'grey'},
    82:{'name':'Gd','color':'grey'},
    83:{'name':'Dy','color':'grey'},
    84:{'name':'Er','color':'grey'},
    85:{'name':'Rh','color':'grey'}
}

def learn_classifier():
    
    training_data_X = np.load('classifier/training_data_X.npy')
    train_Y = np.load('classifier/training_data_Y.npy')

    # initialize train and test sets:
    ## frequent substructures only:
    # train_X = np.zeros((num_molecules*2,num_substructures),bool)
    
    num_samplings = 200
    # 200 = print(len(training_data_X))
    
    classifier_list = []

    for i in range(num_samplings):

        print ('learning',round(((i+1)/float(num_samplings))*100,2),'%')
        
        train_X = training_data_X[i]

        classifier = tree.DecisionTreeClassifier(criterion='entropy', min_samples_leaf=17*3)

        classifier_fit = classifier.fit(train_X, train_Y)

        # print(type(classifier_fit)) : sklearn.tree.tree.DecisionTreeClassifier
        
        classifier_list.append(classifier_fit)

    return classifier_list

def open_dataset(input_file):
    
    # open molecules:
    molecules = pybel.readfile("sdf",input_file)
    
    M_list = []
    M_list_pybel = []
    
    for molecule in molecules:
        G1 = nx.Graph()
        # add vertices:
        for atom in molecule:
            node = atom.idx-1
            element = atom.atomicnum
            G1.add_node(node,label=CHEM_EL[element]['name'],color=CHEM_EL[element]['color'])
        # add edges:
        for bond in openbabel.OBMolBondIter(molecule.OBMol):
            vertex1 = bond.GetBeginAtom().GetIdx()-1
            vertex2 = bond.GetEndAtom().GetIdx()-1
            G1.add_edge(vertex1,vertex2,weight=bond.GetBondOrder())
        # memorize molecule:
        M_list.append(G1)
        M_list_pybel.append(molecule)
  
    return M_list,M_list_pybel

def encode_dataset(M_list, num_substructures, num_samplings):
    """

    Encodes, for each molecule, the subgraphs associated to the classifiers.

    Args:

        M_list (list): 
        num_substructures (int): 100
        num_samplings (int): 200

    Returns:

        A numpy array, with 3 dimensions : classifiers, molecules, 
        substructures.

        Eg. 200 / 4 / 100.

        For each record, we know if the substructure (subgraph) is 
        iso-morphic or not with the molecule.

        If it is iso-morph, this means, the sub-graph is part of the molecule.
        
        It has to be displayed.

    """
    # number of molecules for classification:
    num_molecules_classify = len(M_list)
    
    # dimensions of input data:
    pred_X = np.zeros((num_samplings,num_molecules_classify,num_substructures),bool)
    
    for i in range(num_samplings):
        #UpdateProgressBar(round(((i+1)/float(num_samplings))*100,2))
        print('encoding', round(((i+1)/float(num_samplings))*100,2), '%')
    
        # Open top discriminative substructures for encoding the input data:
        f = open('classifier/top_sg_'+str(i+1)+'.gsp','r')
        SG_list = []
        
        for line in f:
            try:
                if '#' in line:
                    G2 = nx.Graph()
                    line = next(f)
                    while 'v' in line:
                        data = line.split(' ')
                        node = np.int(data[1])
                        element = np.int(data[2])
                        G2.add_node(node, label=CHEM_EL[element]['name'])
                        line = next(f)
                    while 'e' in line:
                        data = line.split(' ')
                        vertex1 = np.int(data[1])
                        vertex2 = np.int(data[2])
                        G2.add_edge(vertex1, vertex2, weight=np.int(data[3]))
                        line = next(f)
                    SG_list.append(G2)
            except ValueError:
                print("Invalid input:", line)
        f.close()
        
        # Encode the input data:
        for i_molecule in range(num_molecules_classify):
            for i_substructure in range(len(SG_list)):
                pred_X[i][i_molecule][i_substructure] = isomorph_test(M_list[i_molecule],SG_list[i_substructure])
                
    return pred_X

def isomorph_test(G1, G2):
    
    em = nx.algorithms.isomorphism.categorical_edge_match('weight', 1)
    nm = nx.algorithms.isomorphism.categorical_node_match('label', 'C')
    GM = nx.algorithms.isomorphism.GraphMatcher(G1,G2,node_match=nm,edge_match=em) #G1-main graph, G2-subgraph
    
    return GM.subgraph_is_isomorphic()

def majority_vote_prediction(clf_list, test_X):
    """This function does the prediction vote for the given classifier list.

Args:
    clf_list (list): the classifier list.
    text_X (numpy.ndarray): a numpy array, with, for each molecule, an array 
    of booleans (the molecule encoded) and a boolean if it is iso-morphic or 
    not with the given substructure.

Returns: 
    A numpy array with, for each molecule (index), True if it is frequent or 
    False if it is non-frequent.

    Only frequent molecules have to be displayed.
    """

    if DEBUG == True:
        print("np.size(test_X,-2) is [" + str(np.size(test_X,-2)) + "]")
        print("len(clf_list) is ["+ str(len(clf_list)) + "]")

    test_results = np.zeros((np.size(test_X,-2),len(clf_list)), bool)
    
    for i in range(len(clf_list)):
        #
        # type(clf_list[i]) : sklearn.tree.tree.DecisionTreeClassifier
        #
        # This actually predicts all input data in one go
        #
        # Updates the value of each of the i-th column array since we have
        # multi-dimensional arrays.
        #
        # We have an array by molecule.
        #
        # Each array length is 200.
        test_results[:,i] = clf_list[i].predict(test_X[i])

    result = np.array([bool(round(np.mean(el))) for el in test_results])

    if DEBUG == True:
        print("result is [" + str(result) + "]")

    return result

def molecule_stat(test_results,test_X,molecule_id):  
    # list of subgraphs which exist in molecule:
    molecule_subgraphs_str = []
    buff_str = ''
    
    for i in range(num_samplings):
        if test_results[molecule_id] == True: # if molecule classified as frequent hitter
            # list of indexes of subgraphs which exist in molecule
            subgraphs_list = [l for l, k in enumerate(test_X[i,molecule_id]) if k == True]
            
            # load the subgraphs
            f = open('classifier/top_sg_'+str(i+1)+'.gsp','r')
            buff_str_list = []
            for line in f:
                try:
                    if '#' in line:
                        while ('#' in line) or ('v' in line) or ('e' in line):
                            line = next(f)
                            buff_str += line
                        buff_str_list.append(buff_str)
                        buff_str = ''
                except ValueError:
                    print("Invalid input:", line)
            f.close()
            
            # form the list of subgraphs
            for subgraph in subgraphs_list:
                molecule_subgraphs_str.append(buff_str_list[subgraph])
    #%
    top_sg_list = []
    subgraphs_frequencies = []
    
    for graph,freq in collections.Counter(molecule_subgraphs_str).most_common():
        top_sg_list.append('#\n'+graph)
        subgraphs_frequencies.append(freq)
        
    #SG_list = []
    Mol_list = []
    for line in top_sg_list:
        try:
            if '#' in line:
                line_lst = line.rstrip().split('\n')
                #G3 = nx.Graph()
                mol = openbabel.OBMol()
                for entry in line_lst:
                    if 'v' in entry:
                        data = entry.split(' ')
                        #node = np.int(data[1])
                        element = np.int(data[2])
                        #G3.add_node(node,label=CHEM_EL[element]['name'],color=CHEM_EL[element]['color'])
                        a = mol.NewAtom()
                        a.SetAtomicNum(element)
                    elif 'e' in entry:
                        data = entry.split(' ')
                        vertice1 = np.int(data[1])
                        vertice2 = np.int(data[2])
                        #G3.add_edge(vertice1,vertice2,weight=np.int(data[3]))
                        mol.AddBond(vertice1+1, vertice2+1, np.int(data[3]))
                #SG_list.append(G3)
                pybelMol = pybel.Molecule(mol)
                Mol_list.append(pybelMol)
        except ValueError:
            print("Invalid input:", line)
    
    return Mol_list,subgraphs_frequencies #SG_list,subgraphs_frequencies

def my_molecule_stat(clf_list, test_results, test_X, molecule_id):
    """This method is a new implementation of the molecule_stat method.

    Args:
        clf_list(list): the classifier list (length 200 normally).
        test_results (list): a boolean array with, for each molecule, if it has to be displayed or not.
        test_X (numpy.ndarray): The encoded data set numpy array (eg. 200 / 4 / 100).
        molecule_id: the identifier of the molecule we are processing.

    Results:
        The list of the subgraphs and their frequencies.

    """

    for i in range(num_samplings):
        
        if test_results[molecule_id] == True: 
            # molecule is classified as frequent hitter

            subgraphs_indexes_list = [l for l, k in enumerate(test_X[i,molecule_id]) if k == True]

            if DEBUG == True:
                print("subgraphs_indexes_list is ["+str(subgraphs_indexes_list)+"]")

            node_indicator = clf_list[i].decision_path(test_X[i])

            print(node_indicator)
            print(node_indicator.todense())
            print(node_indicator.indices[node_indicator.indptr[0]: node_indicator.indptr[1]])

            leave_id = clf_list[i].apply(test_X[i])

            print (leave_id)

            #
            # begin of the code copy pasted
            #
            # source : https://scikit-learn.org/stable/auto_examples/tree/plot_unveil_tree_structure.html
            #

            feature = clf_list[i].tree_.feature
            threshold = clf_list[i].tree_.threshold

            # node_indicator = clf_list[i].decision_path(test_X[i])

            # Similarly, we can also have the leaves ids reached by each sample.

            leave_id = clf_list[i].apply(test_X[i])

            # Now, it's possible to get the tests that were used to predict a sample or
            # a group of samples. First, let's make it for the sample.

            sample_id = 0
            node_index = node_indicator.indices[node_indicator.indptr[sample_id]:
                                                node_indicator.indptr[sample_id + 1]]

            print('Rules used to predict sample %s: ' % sample_id)
            for node_id in node_index:
                if leave_id[sample_id] == node_id:
                    continue

                if (test_X[i][sample_id][feature[node_id]] <= threshold[node_id]):
                    threshold_sign = "<="
                else:
                    threshold_sign = ">"

                print("decision id node %s : (X_test[%s, %s] (= %s) %s %s)"
                    % (node_id,
                        sample_id,
                        feature[node_id], # this one, for the classifiers.
                        test_X[i][sample_id][feature[node_id]], # this is the test
                        threshold_sign,
                        threshold[node_id]))

                if test_X[i][sample_id][feature[node_id]]:

                    if DEBUG:
                        print("Ok")
                        print(feature[node_id])

            #
            # End of the code copy pasted
            #

            file = open('classifier/top_sg_'+str(i+1)+'.gsp','r')

            buff_str = ''

            buff_str_list = []
            for line in file:
                try:
                    if '#' in line:
                        while ('#' in line) or ('v' in line) or ('e' in line):
                            line = next(file)
                            buff_str += line
                        buff_str_list.append(buff_str)
                        buff_str = ''
                except ValueError:
                    print("Invalid input:", line)
            file.close()

            # if DEBUG == True:
            #    print (buff_str_list)

            # 
            # Alors, maintenant, il faut récupérer les sous-graphes et les 
            # confronter aux arbres de décisions.
            #

            for subgraph_index in subgraphs_indexes_list:
                if DEBUG == True:
                    # print ("buff_str_list is ["+str(buff_str_list[subgraph_index])+"]")
                    pass

    sys.exit(0)

    return [], []

if __name__ == "__main__":
    
    classifier_list = learn_classifier()
    
    # print (str(classifier_list))
    print (len(classifier_list))
    
    file = "demo.sdf"

    if (len(sys.argv) > 2):
        print(str(sys.argv))
        print("too many arguments")
        sys.exit(1)
    elif (len(sys.argv) == 2):
        file = sys.argv[1]

    if DEBUG == True:
        print("processing file [" + str(file) + "]")

    # on va ouvrir un fichier avec la méthode open_dataset.
    M_dataset, M_compounds = open_dataset(file)

    print(len(M_dataset))   # => 1 seule molecule dans demo.sdf
    print(len(M_compounds)) # => 1 seul graphe, puisqu'une seule molecule dans demo.sdf

    #
    # len(M_dataset) = len(M_compounds) 
    #
    
    #
    # M_dataset => Liste de graphes (noeuds / atomes et arrêtes).
    #
    # https://networkx.github.io/documentation/networkx-1.10/tutorial/tutorial.html
    #
    # M_compounds => Liste de molecules pybel
    #
    # http://openbabel.org/docs/current/UseTheLibrary/Python_PybelAPI.html#pybel.Molecule
    #

    num_substructures = 100
    num_samplings = 200
    # 200 = print(len(training_data_X)) = print(len(classifier_list))
    
    dataset_X = encode_dataset(M_dataset, num_substructures, num_samplings)
    
    if DEBUG == True:
        print(dataset_X)
        print(dataset_X[0])
        print(dataset_X[0][0])

    # pred_res = np.zeros(len(M_dataset),bool)
    pred_res = majority_vote_prediction(classifier_list, dataset_X)
    
    for molecule_id in range(len(M_compounds)):

        graphs_list, graphs_freq = my_molecule_stat(classifier_list, pred_res, dataset_X, molecule_id)

        print (graphs_list)
        print (graphs_freq)
        
        print (len(graphs_list))
        print (len(graphs_freq))

        for molecule in graphs_list:
            print(molecule)
    
    sys.exit(0)
